/* generates frames for PC's raytracer,Stig A. Olsen, 1995 (stigo@ifi.uio.no)
   usage: color > color.dat
   or
   usage: color <#n> <#t> > framename<#n>.dat

   the last generates frame number n of a total of t frames,
   e.g.
   color 3 16 > bogus003.dat
   generates frame number 3 of 16.
*/

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>

#define POINTS 7

#define RADIUS 100.0
#define q 40.0

char *materials[]={
"#material",
"mat1, 0.5, 0.5, 0.5,  0.4, 0.9, 0.8,  100.0,",
"#material",
"mat2, 0.35, 0.35, 0.35,  0.2, 0.9, 0.0,  200.0,",
"none",
};

int main(int argc, char **argv) {
  double x, y, z, ny, nx, nz, counter, offset, size, delta;
  int indeks, i; 
  float frames, frame;
  char *ptr;
  char *lineptr;

  if(argc!=1) {
    frame=atof(argv[1]);
    frames=atof(argv[2]);

    offset=(2*PI/POINTS)/frames*frame;
    delta=(2*PI/POINTS)/2;
  }
  else offset=0.0;

  /* print standard information */
  /* bgcol(r,g,b), ambli lys(r,g,b), eye_range */
  printf("#scene\n 0.0, 0.0, 0.0,  1.0, 1.0, 1.0,  2000.0,\n");
  /* src(x,y,z), dst(x,y,z) */

  printf("400.0, 400.0, -800.0,   0.0, 0.0, 0.0,\n");
  printf("0.0, 1.0,\n1000, 10, 320, 200, 1,\n");

  i=0;
  ptr=materials[i];
  while(strcmp(ptr, "none")) {
    printf("%s\n", ptr);
    i++;
    ptr=materials[i];
  }

  printf("#light\n");
  printf("302.0, 30.0,  -30.0,  0.3, 0.3, 0.9,  50000.0,\n");
  printf("#light\n");
  printf("30.0, 302.0,  -30.0,  0.3, 0.9, 0.9,  50000.0,\n");
  printf("#light\n");
  printf("20.0, 0.0,  -302.0,  0.7, 0.3, 0.9,  50000.0,\n");

  /* center sphere */
  printf("#sphere\n");
  printf("mat2, 0,0,0, 75,\n"); 
  size=35.0;
  for(i=0; i<POINTS; i++) {
    counter=(2*PI/POINTS)*i;
    x=(RADIUS*1.4)*cos(counter+offset);
    z=RADIUS*sin(counter+offset);
    y=0.0;
    size=35.0+abs(10*cos(counter+offset));
    printf("#sphere\n");
    printf("mat1, %f,%f,%f, %f,\n", x,y,z,size);
    y=RADIUS*cos(counter+offset+delta);
    z=RADIUS*sin(counter+offset+delta);
    size=35.0;
    x=0.0;
    printf("#sphere\n");
    printf("mat1, %f,%f,%f, %f,\n", x,y,z,size);
  }
}

