// Copyright (c) 1994,1995 by Per Christian Oedegaard.  All Rights Reserved
// ***********************************************
// "Speed" - Triumph intro released at TG95
// ***********************************************
// Code:  Per "Quark" degrd
// Heavy patching and optimizations by
// Nils "Darkman" Corneliusen
// ***********************************************
// Part:  Tracer
// ***********************************************
// PRELIMINARY AND CONFIDENTIAL!
// ***********************************************


#include <math.h>
#include <m68881.h>


__far __asm extern void asmHitMe( void );         // from start.s
__far __asm extern void entf1( void );            // from start.s
__far __asm extern void entf2( void );            // from start.s
__far __asm extern void entf3( void );            // from start.s
__far __asm extern void entf4( void );            // from start.s
__far __asm extern void entf5( void );            // from start.s
__far __asm extern void entf6( void );            // from start.s
__far __asm extern void ents1( void );            // from start.s
__far extern char *DestPtr;
__far extern char *ScenePtr;

#define TRUE  1
#define FALSE 0

// enum command {SCENE, LIGHT, TRIANGLE, SPHERE, MATERIAL, UNDEF};
#define SCENE    999
#define LIGHT    1000
#define TRIANGLE 1001
#define SPHERE   1002
#define MATERIAL 1003
#define UNDEF    1004

#define command int

#define dot(x1,y1,z1,x2,y2,z2) (x1*x2 + y1*y2 + z1*z2)

enum hitmode {DIFFUSE, REFLECT, BACKGROUND};


// Normalizes the triple (a,b,c)

__inline void normalize(double& a, double& b, double& c)
{
  double len;
  len = sqrt(a*a + b*b + c*c);
  a = a/len; b = b/len; c = c/len;
}




//
// ******** STRUCTURES AND CLASSES ********
//


struct bounding_box                         // Bounding box structure
{
  double min_x, max_x, min_y, max_y, min_z, max_z;
};


struct a_scene
{
	double bgcol_r, bgcol_g, bgcol_b;         // Background color
	double ambli_r, ambli_g, ambli_b;         // Ambient light intensity
	double eye_rng;
	double src_x, src_y, src_z;
	double dest_x, dest_y, dest_z;
	double bank, zoom;
	int max_iter;                             // Max # of beam-iterations
	int max_depth;                            // Max depth of ray-tree
	int width, height, aaval;                 // Picture size
	bounding_box bbox;

	void set( char *src )
	{
		bgcol_r = *((double *)src); src+=8;
		bgcol_g = *((double *)src); src+=8;
		bgcol_b = *((double *)src); src+=8;
		ambli_r = *((double *)src); src+=8;
		ambli_g = *((double *)src); src+=8;
		ambli_b = *((double *)src); src+=8;
		eye_rng = *((double *)src); src+=8;
		src_x = *((double *)src); src+=8;
		src_y = *((double *)src); src+=8;
		src_z = *((double *)src); src+=8;
		dest_x = *((double *)src); src+=8;
		dest_y = *((double *)src); src+=8;
		dest_z = *((double *)src); src+=8;
		bank = *((double *)src); src+=8;
		zoom = *((double *)src); src+=8;

		max_iter  = *((int *)src); src+=4;
		max_depth = *((int *)src); src+=4;
		width     = *((int *)src); src+=4;
		height    = *((int *)src); src+=4;
		aaval     = *((int *)src); src+=4;

		bbox.min_x = bbox.max_x = src_x;
		bbox.min_y = bbox.max_y = src_y;
		bbox.min_z = bbox.max_z = src_z;

	}

} scene;

struct info_block
{
  double x, y, z, dx, dy, dz, r, g, b, factor;
  hitmode mode;
} info, tmpinfo;

class material                          // Material class
{
public:
  int name;
  double red, green, blue;
  double k_amb, k_diff, k_spec, spot;

  material *next;
  material() {next = 0;}

  void set(char *src)
  {
		name = *((int *)src);
		src+=8;

		red    = *((double *)src); src+=8;
		green  = *((double *)src); src+=8;
		blue   = *((double *)src); src+=8;
		k_amb  = *((double *)src); src+=8;
		k_diff = *((double *)src); src+=8;
		k_spec = *((double *)src); src+=8;
		spot   = *((double *)src); src+=8;

  }
} *mat_head = 0;


class lightsrc                          // Lightsource class
{
public:
	double pos_x, pos_y, pos_z;           // Lightsource position
	double red, green, blue;              // Its color
	double intensity;                     // Its intensity

	lightsrc *next;                       // Pointer to next lightsource
	lightsrc() {next = 0;}
	void set( char *src )
	{

		pos_x     = *((double *) src ); src+=8;
		pos_y     = *((double *) src ); src+=8;
		pos_z     = *((double *) src ); src+=8;
		red       = *((double *) src ); src+=8;
		green     = *((double *) src ); src+=8;
		blue      = *((double *) src ); src+=8;
		intensity = *((double *) src ); src+=8;

	}
} *light_head = 0;






// *** GEO ***
// The class 'geo' contains the common elements of all
// geometric shapes to appear in the scene.

class geo *geo_head = 0;

class geo
{
  double pos_x, pos_y, pos_z;               // The object's position
  double nx, ny, nz;
  material *mat;                            // The object's material
  bounding_box bb;                          // The object's bounding box  
  friend class triangle;
  friend class sphere;

public:
  geo *next;
  geo() {next = 0;}                         // Constructor: No succeeding object at the moment
  virtual void set(char *, material *) = 0;

  void get_bbox( double& mi_x, double& ma_x, double& mi_y, double& ma_y, double& mi_z, double& ma_z )
  {
    mi_x = bb.min_z; ma_x = bb.max_x; mi_y = bb.min_y;
    ma_y = bb.max_y; mi_z = bb.min_z; ma_z = bb.max_z;
  }

  virtual void getns(double, double, double) = 0;

  virtual int intersect(double, double, double, double, double, double, double&) = 0;

  void trace(double& zval)
  {
    double t, tt;                           // The affine solution
    double x, y, z;                         // Intersection point and normal vector
    double dx, dy, dz;
    double lx, ly, lz, nl, len, lilen;
    double rdot, rint;
    double specdot;
    int hit;
    lightsrc *li;
    geo *ge;

    if( intersect(info.x, info.y, info.z, info.dx, info.dy, info.dz, t) ) {
      x = t*info.dx; y = t*info.dy; z = t*info.dz;
      len = x*x + y*y + z*z;                // Sqrt omitted for speed
      x += info.x; y += info.y; z += info.z;

      if((len < zval) && (t > 0.000000001)) {
        zval = len;
        getns(x, y, z);                     // Normal vectors

        if(mat->k_spec != 0.0) {
          tmpinfo.factor = mat->k_spec;
          tmpinfo.x = x; tmpinfo.y = y; tmpinfo.z = z;
          tmpinfo.mode = REFLECT; 
          dx = info.dx; dy = info.dy; dz = info.dz;
          normalize(dx,dy,dz);
          rdot = 2*dot(nx,ny,nz, dx,dy,dz);
//        if(rdot<0) rdot=0;
          tmpinfo.dx = -(rdot*nx - dx);     // Optimized
          tmpinfo.dy = -(rdot*ny - dy);
          tmpinfo.dz = -(rdot*nz - dz);
        } else {
          tmpinfo.mode = DIFFUSE;
          tmpinfo.factor = 1.0;
        }

        // Ambient light
        tmpinfo.r = scene.ambli_r * mat->k_amb * mat->red;
        tmpinfo.g = scene.ambli_g * mat->k_amb * mat->green;
        tmpinfo.b = scene.ambli_b * mat->k_amb * mat->blue;

        // Adding light-source induced colour
        li = light_head;

        while(li != 0) {
          lx = li->pos_x - x;
          ly = li->pos_y - y;
          lz = li->pos_z - z;
          lilen = lx*lx + ly*ly + lz*lz;    // Squared length to lightsource

          ge = geo_head;                    // Objects shadowing?
          hit = FALSE;

          while((ge != 0 && !hit)) {
            if(ge != this) {
              hit = ge->intersect(x,y,z, lx,ly,lz, tt);
              if((tt < 0.0) || (tt > 1.0)) hit = FALSE;
            }
            ge = ge->next;
          }

          if(!hit) {

            // was: normalize(lx, ly, lz), this saves a*a+b*b+c*c + automatic inline :)
            specdot = sqrt(lilen);
            lx = lx/specdot;
            ly = ly/specdot;
            lz = lz/specdot;
            // end


            nl = 2 * dot(nx,ny,nz, lx,ly,lz);
            if(nl < 0.0) nl = 0;

            dx = x; dy = y; dz = z + scene.eye_rng;
            normalize(dx,dy,dz);

            lx = -(nl*nx - lx);
            ly = -(nl*ny - ly);
            lz = -(nl*nz - lz);

            rdot = dot(dx,dy,dz, lx,ly,lz);
            if (rdot < 0.0) rdot = 0.0; else rdot = pow(rdot,mat->spot);

            rint = li->intensity / lilen;   // Relative intensity
            specdot = mat->k_spec*rdot;  // optimized

            tmpinfo.r += rint * li->red *   ( mat->k_diff * mat->red   * nl + specdot );
            tmpinfo.g += rint * li->green * ( mat->k_diff * mat->green * nl + specdot );
            tmpinfo.b += rint * li->blue *  ( mat->k_diff * mat->blue  * nl + specdot );
          }
          li = li->next;
        }
      }
    }
  }
};

// The subclass triangle contains the elements neccessary
// to describe a triangle.

class triangle:public geo
{
  double x1,y1,z1, x2,y2,z2, x3,y3,z3;
  double nd;

public:

	void exg(double& a1, double& a2, double& b1, double& b2)
	{
		double tmp;
		tmp = a1; a1 = a2; a2 = tmp;
		tmp = b1; b1 = b2; b2 = tmp;
	}

  void exg(double& a1, double& a2)
  {
    double tmp;
    tmp = a1; a1 = a2; a2 = tmp;    
  }

  // optimize max/min usage???

  void set(char *src, material*mt)
  {
    double max, min;
    double v1, v2, v3, w1, w2, w3;

    mat = mt;

	pos_x = *((double *)src);	src+=8;
	pos_y = *((double *)src);	src+=8;
	pos_z = *((double *)src);	src+=8;

	x1 = *((double *)src);	src+=8;
	y1 = *((double *)src);	src+=8;
	z1 = *((double *)src);	src+=8;
	x2 = *((double *)src);	src+=8;
	y2 = *((double *)src);	src+=8;
	z2 = *((double *)src);	src+=8;
	x3 = *((double *)src);	src+=8;
	y3 = *((double *)src);	src+=8;
	z3 = *((double *)src);	src+=8;

    x1+=pos_x; x2+=pos_x; x3+=pos_x;
    y1+=pos_y; y2+=pos_y; y3+=pos_y;
    z1+=pos_z; z2+=pos_z; z3+=pos_z;

    // Sort in increasing y order

    min = x1; if(x2 < min) min = x2; if(x3 < min) min = x3;
    max = x1; if(x2 > max) max = x2; if(x3 > max) max = x3;
    bb.min_x = min; bb.max_x = max;
    if(min < scene.bbox.min_x) scene.bbox.min_x = min;
    if(max > scene.bbox.max_x) scene.bbox.max_x = min;

    min = y1; if(y2 < min) min = y2; if(y3 < min) min = y3;
    max = y1; if(y2 > max) max = y2; if(y3 > max) max = y3;
    bb.min_y = min; bb.max_y = max;
    if(min < scene.bbox.min_y) scene.bbox.min_y = min;
    if(max > scene.bbox.max_y) scene.bbox.max_y = min;

    min = z1; if(z2 < min) min = z2; if(z3 < min) min = z3;
    max = z1; if(z2 > max) max = z2; if(z3 > max) max = z3;
    bb.min_z = min; bb.max_z = max;
    if(min < scene.bbox.min_z) scene.bbox.min_z = min;
    if(max > scene.bbox.max_z) scene.bbox.max_z = min;

    v1 = x2 - x1; v2 = y2 - y1; v3 = z2 - z1;
    w1 = x3 - x1; w2 = y3 - y1; w3 = z3 - z1;
    nx = v2*w3 - v3*w2; ny = v3*w1 - v1*w3; nz = v1*w2 - v2*w1;
    normalize(nx,ny,nz);
    nd = -(nx*x1 + ny*y1 + nz*z1);          // Plane eq. calculated
  }

  virtual void getns(double, double, double) {};

  virtual int intersect(double x0, double y0, double z0, double dx, double dy, double dz, double& t)
  {

    double de, nxd,nyd,nzd;
    double px, py, pz;
    double lam, my, v1,v2, w1,w2, det;
    double b1,b2,b3,b4, del1, del2;
    int hit_state = FALSE;

    de = (nx*dx + ny*dy + nz*dz);

    if(de != 0.0) {
      t = -(nx*x0 + ny*y0 + nz*z0 + nd) / de;
      px = x0 + t*dx;
      py = y0 + t*dy;
      pz = z0 + t*dz;

      nxd = fabs(nx); nyd = fabs(ny); nzd = fabs(nz);

      if((nzd > nxd) && (nzd > nyd)) {
        v1 = x2-x1; w1 = x3-x1;                // Projection onto the XY-plane
        v2 = y2-y1; w2 = y3-y1;
        del1 = px-x1; del2 = py-y1;
      } else {
        if((nyd > nxd) && (nyd > nxd)) {
          v1 = x2-x1; w1 = x3-x1;              // Projection onto the XZ-plane
          v2 = z2-z1; w2 = z3-z1;
          del1 = px-x1; del2 = pz-z1;
        } else {
          v1 = y2-y1; w1 = y3-y1;              // Projection onto the YZ-plane
          v2 = z2-z1; w2 = z3-z1;
          del1 = py-y1; del2 = pz-z1;
        }
      }

      det = v1*w2 - w1*v2;                     // moved out (save some code :)

      // removed det check here! (nilsco)

      det = 1/det;
      b1 = w2*det; b2 = -w1*det; b3 = -v2*det; b4 = v1*det;
      lam = del1*b1 + del2*b2;
      my  = del1*b3 + del2*b4;

      if((lam >= 0.0) && (my >= 0.0)  && ((lam+my) <= 1.0)) {
        hit_state = TRUE;
        if(dot(nx,ny,nz, dx,dy,dz) > 0.0) {
          nx=-nx; ny=-ny; nz=-nz; nd = -(nx*x1 + ny*y1 + nz*z1);
        }
      }
    }
    return(hit_state);
  }
};

// The subclass sphere contains the elements neccessary
// to describe a sphere.

class sphere:public geo
{
  double radius;

public:
  virtual void set(char *src, material*mt)
  {
    mat = mt;

	pos_x = *((double *)src);	src+=8;
	pos_y = *((double *)src);	src+=8;
	pos_z = *((double *)src);	src+=8;
	radius = *((double *)src);	src+=8;

    bb.min_x = pos_x - radius; bb.max_x = pos_x + radius;
    bb.min_y = pos_y - radius; bb.max_y = pos_y + radius;
    bb.min_z = pos_z - radius; bb.max_z = pos_z + radius;

    if(bb.min_x < scene.bbox.min_x) scene.bbox.min_x = bb.min_x;
    if(bb.max_x > scene.bbox.max_x) scene.bbox.max_x = bb.max_x;
    if(bb.min_y < scene.bbox.min_y) scene.bbox.min_y = bb.min_y;
    if(bb.max_y > scene.bbox.max_y) scene.bbox.max_y = bb.max_y;
    if(bb.min_z < scene.bbox.min_z) scene.bbox.min_z = bb.min_z;
    if(bb.max_z > scene.bbox.max_z) scene.bbox.max_z = bb.max_z;
  }

  virtual void getns(double x, double y, double z)
  {
    nx = (x - pos_x)/radius;
    ny = (y - pos_y)/radius;
    nz = (z - pos_z)/radius;
    normalize(nx, ny, nz);
  }

  virtual int intersect(double x0, double y0, double z0, double dx, double dy, double dz, double& t)
  {
    int hit_state = FALSE;
    double a, b;
    double t1, t2;                          // The affine solutions
    double ex, sq;                          // Help variables

    // optimized (nilsco)
    double ztx,zty,ztz;
    double ztl,ztm;

    ztx = x0-pos_x;
    zty = y0-pos_y;
    ztz = z0-pos_z;

    a = dx*dx + dy*dy + dz*dz;
    b = 2*( dx*ztx + dy*zty + dz*ztz );

    ztl = a * (ztx*ztx + zty*zty + ztz*ztz - radius*radius);  // moved c in here (nilsco)
    ztm = a+a;

    ex = b*b - 4*ztl;
    if((ex > 0) && (ztl != 0)) {
      hit_state = TRUE;                     // Intersected
      sq = sqrt(ex);
      t1 = (-b + sq) / ztm;
      t2 = (-b - sq) / ztm;
      t = t1; if((t2 < t1) && (t2 > 0)) t = t2;
      if(t < 0.0) hit_state = FALSE;
    }
    return(hit_state);
  }
};



material *getmat(char *castme)
{
	material *curmat = mat_head;
	int found = FALSE;

	while(!found)
	{
		if( *((int *)castme) == curmat->name ) found = TRUE;
		else curmat = curmat->next;
	}

	return(curmat);
}


// Builds structures

void init_struct( char *presc )
{

	lightsrc *curli = light_head, *li = 0;
	geo *curge = geo_head, *ge = 0;
	material *curmat = mat_head, *mat = 0;

	char *tmpbuf2;
	int comm;

	tmpbuf2 = presc;

	while( (comm = *((int *) tmpbuf2)) != -1 ) {

		tmpbuf2+=8;

		switch(comm) {

		case SCENE:
			scene.set( tmpbuf2 ) ;
			tmpbuf2+=(sizeof(double)*15+sizeof(int)*6);

			break;

		case LIGHT:
			li = new lightsrc;
			if(light_head == 0) light_head = li; else curli->next = li;

			li->set( tmpbuf2 );
			tmpbuf2+=(sizeof(double)*7);

			curli = li;
			break;


		case TRIANGLE:
			ge = new triangle;
			if(geo_head == 0) geo_head = ge; else curge->next = ge;

			mat = getmat(tmpbuf2);
			tmpbuf2+=8;

			ge->set(tmpbuf2,mat);
			tmpbuf2+=sizeof(double)*12;

			curge = ge;
			break;


		case SPHERE:
			ge = new sphere;
			if(geo_head == 0) geo_head = ge; else curge->next = ge;

			mat = getmat(tmpbuf2);
			tmpbuf2+=8;

			ge->set(tmpbuf2,mat);
			tmpbuf2+=sizeof(double)*4;

			curge = ge;
			break;


		case MATERIAL:
			mat = new material;
			if(mat_head == 0) mat_head = mat; else curmat->next = mat;

			mat->set(tmpbuf2);
			tmpbuf2+=(sizeof(int)*2+sizeof(double)*7);

			curmat = mat;
			break;

		}
	}

}



void massacre()
{
  geo *curgeo = geo_head, *nextgeo;
  lightsrc *curli = light_head, *nextli;
  material *curmat = mat_head, *nextmat;

  while(curgeo != 0)                    // Deleting geometric objects
  {
    nextgeo = curgeo->next;
    delete curgeo;
    curgeo = nextgeo;
  }
  geo_head = 0;

  while(curli != 0)                     // Deleting light-sources
  {
    nextli = curli->next;
    delete curli;
    curli = nextli;
  }
  light_head = 0;

  while(curli != 0)                     // Deleting materials
  {
    nextmat = curmat->next;
    delete curmat;
    curmat = nextmat;
  }
  mat_head = 0;
}


//
// *** THE ACTUAL RAYTRACING ***
//


void trace( char *mem_dest )
{
  int bounce, rtms, max_rtms=1;             // Coordinates
  double x1, y1, z1, dx, dy, dz;
  double i,j, fact, zval, bigval;
  double eye_x, eye_y, eye_z, rx, ry, rz, tx, ty, tz, wh, hh;
  double drx, dry, drz, dtx, dty, dtz;
  double red, green, blue ;

  unsigned char *copyptr;

  geo *ge;

  unsigned char *temppict;

  temppict = (unsigned char *) mem_dest;

  dx = scene.dest_x - scene.src_x;
  dy = scene.dest_y - scene.src_y;
  dz = scene.dest_z - scene.src_z;
  normalize(dx,dy,dz);

  eye_x = scene.src_x - scene.eye_rng*dx;
  eye_y = scene.src_y - scene.eye_rng*dy;
  eye_z = scene.src_z - scene.eye_rng*dz;

  wh = (double)scene.width / 2.0;
  hh = (double)scene.height / 2.0;

  // Obscure formula ahead :-)

  rx = dz; ry = 0.0; rz = -dx;
  if((rx == 0.0) && (rz == 0.0)) { rx = 1.0; rz = 0.0; }
  else normalize(rx, ry, rz);

  // Up-vector is view-vector crossed with right-vector :-?

  tx = dy*rz - dz*ry; ty = dz*rx - dx*rz; tz = dx*ry - dy*rx;
  
  // zoom is removed here!
  // banking removed! (avoids use of cos/sin) cosb=1, sinb=0

  drx = rx; dry = ry; drz = rz;
  dtx = tx; dty = ty; dtz = tz;


//  normalize(drx,dry,drz); normalize(dtx,dty,dtz);
  rx = -wh*drx + hh*dtx + scene.src_x;
  ry = -wh*dry + hh*dty + scene.src_y;
  rz = -wh*drz + hh*dtz + scene.src_z;

      bigval =                              // Greatest possible length
        (scene.bbox.max_x - scene.bbox.min_x + scene.eye_rng) *
        (scene.bbox.max_x - scene.bbox.min_x + scene.eye_rng) +
        (scene.bbox.max_y - scene.bbox.min_y + scene.eye_rng) *
        (scene.bbox.max_y - scene.bbox.min_y + scene.eye_rng) +
        (scene.bbox.max_z - scene.bbox.min_z + scene.eye_rng) *
        (scene.bbox.max_z - scene.bbox.min_z + scene.eye_rng);

  for(j = 0; j < (double)scene.height; j+=2) {
    copyptr = temppict;

    for(i = 0; i < (double)scene.width; i+=2) {

      red = green = blue = 0.0;

      info.x = eye_x; info.y = eye_y; info.z = eye_z;
      x1 = rx + i*drx - j*dtx;
      y1 = ry + i*dry - j*dty;
      z1 = rz + i*drz - j*dtz;
      info.dx = x1 - info.x; info.dy = y1 - info.y;
      info.dz = z1 - info.z; info.r = 0.0; info.g = 0.0; info.b = 0.0;

      fact = 1.0;
      rtms = 1;

      do {
        ge = geo_head;
        zval = bigval;

        bounce = FALSE;
        tmpinfo.r = scene.bgcol_r; tmpinfo.g = scene.bgcol_g;
        tmpinfo.b = scene.bgcol_b; tmpinfo.mode = BACKGROUND;

        while(ge != 0) {
          ge->trace(zval);
          ge=ge->next;
        }

        if(tmpinfo.mode == REFLECT) {
          info.x = tmpinfo.x; info.y = tmpinfo.y; info.z = tmpinfo.z;
          info.dx = tmpinfo.dx; info.dy = tmpinfo.dy; info.dz = tmpinfo.dz;
          info.r += fact*tmpinfo.r;
          info.g += fact*tmpinfo.g;
          info.b += fact*tmpinfo.b;
          fact *= tmpinfo.factor;           // Factor for next reflection
          rtms += 1;
          if(rtms <= scene.max_iter) bounce = TRUE;
        } else {
          info.r += fact*tmpinfo.r;
          info.g += fact*tmpinfo.g;
          info.b += fact*tmpinfo.b;
        }

      } while (bounce);

      if(rtms > max_rtms) max_rtms = rtms;
      red += info.r; green +=info.g; blue +=info.b;

      if(red   > 1.0) red = 1.0;
      if(green > 1.0) green = 1.0;
      if(blue  > 1.0) blue = 1.0;

      *temppict++ = ((unsigned char)(255*red));
      *temppict++ = ((unsigned char)(255*green));
      *temppict++ = ((unsigned char)(255*blue));
      *temppict++ = ((unsigned char)(255*red));
      *temppict++ = ((unsigned char)(255*green));
      *temppict++ = ((unsigned char)(255*blue));
    }

    for( int z=0; z < scene.width*3; z++ )
      *temppict++ = *copyptr++;

  }
}



extern void traceme( char *desc, char *mem_dest )
{
	init_struct( desc );
	trace( mem_dest );
	massacre();
}

int main( void )
{
	asmHitMe();
	traceme( ScenePtr, DestPtr );
	entf1();
	traceme( ScenePtr, DestPtr );
	entf1();
	traceme( ScenePtr, DestPtr );
	entf1();
	traceme( ScenePtr, DestPtr );
	entf1();
	traceme( ScenePtr, DestPtr );
	entf1();
	traceme( ScenePtr, DestPtr );
	entf1();
	traceme( ScenePtr, DestPtr );
	entf1();
	traceme( ScenePtr, DestPtr );
	entf1();
	traceme( ScenePtr, DestPtr );
	entf2();
	traceme( ScenePtr, DestPtr );
	entf2();
	traceme( ScenePtr, DestPtr );
	entf2();
	traceme( ScenePtr, DestPtr );
	entf2();
	traceme( ScenePtr, DestPtr );
	entf3();
	traceme( ScenePtr, DestPtr );
	entf3();
	traceme( ScenePtr, DestPtr );
	entf3();
	traceme( ScenePtr, DestPtr );
	entf3();
	traceme( ScenePtr, DestPtr );
	entf4();
	traceme( ScenePtr, DestPtr );
	entf4();
	traceme( ScenePtr, DestPtr );
	entf4();
	traceme( ScenePtr, DestPtr );
	entf4();
	traceme( ScenePtr, DestPtr );
	entf5();
	traceme( ScenePtr, DestPtr );
	entf5();
	traceme( ScenePtr, DestPtr );
	entf6();
	traceme( ScenePtr, DestPtr );
	entf6();
	traceme( ScenePtr, DestPtr );
	ents1();

	return 0;
}
